﻿using System;
using Ozeki.Media;
using Ozeki.MediaGateway;
using Ozeki.VoIP;

namespace Web2SIPServer
{
    class MyClient
    {
        // Client reference
        IClient _client;

        ISoftPhone _softPhone;
        IPhoneLine _phoneLine;
        IPhoneCall _call;

        IStreamService _streamService;
        string _publishStreamName;
        string _playStreamName;

        MediaConnector _mediaConnector;
        PhoneCallAudioSender _phoneCallAudioSender;
        PhoneCallAudioReceiver _phoneCallAudioReceiver;

        MediaGatewayAudioReceiver _mediaGatewayAudioReceiver;
        MediaGatewayAudioSender _mediaGatewayAudioSender;

        /// <summary>
        /// Creates a new instance of MyClient.
        /// </summary>
        /// <param name="client">The connected client.</param>
        /// <param name="streamService"> Stream service what transmits media data between two client.</param>
        /// <param name="softPhone"> Represents a SIP softphone</param>
        /// <param name="account">SIP account for PBX registration.</param>
        public MyClient(IClient client, IStreamService streamService, ISoftPhone softPhone, SIPAccount account)
        {
            _mediaConnector = new MediaConnector();
            _phoneCallAudioSender = new PhoneCallAudioSender();
            _phoneCallAudioReceiver = new PhoneCallAudioReceiver();

            _streamService = streamService;
            _softPhone = softPhone;
            _client = client;

            Account = account;
        }

        /// <summary>
        /// SIP account for able to register a SIP PBX.
        /// </summary>
        public SIPAccount Account { get; private set; }

        /// <summary>
        /// Registers a SIP account for a SIP PBX.
        /// </summary>
        public virtual void Register()
        {
            _phoneLine = _softPhone.CreatePhoneLine(Account);

            _phoneLine.RegistrationStateChanged += PhoneLine_RegistrationStateChanged;
            _softPhone.IncomingCall += SoftPhone_IncomingCall;

            _softPhone.RegisterPhoneLine(_phoneLine);
        }

        /// <summary>
        /// Unregisters the registered phoneline from the SIP PBX. 
        /// </summary>
        public virtual void Unregister()
        {
            if (_mediaGatewayAudioReceiver != null)
            {
                _streamService.RemoveStream(_mediaGatewayAudioReceiver.MediaStream);
                _mediaGatewayAudioReceiver = null;
            }

            HangUp();

            _softPhone.IncomingCall -= SoftPhone_IncomingCall;

            if (_phoneLine != null)
            {
                _phoneLine.RegistrationStateChanged -= PhoneLine_RegistrationStateChanged;
                _softPhone.UnregisterPhoneLine(_phoneLine);
            }
        }

        /// <summary>
        /// Call the given dial number.
        /// </summary>
        /// <param name="dialNumber">dial number</param>
        public virtual void Call(string dialNumber)
        {
            if (_phoneLine == null)
            {
                System.Diagnostics.Debug.Fail("PhoneLine null");
                return;
            }

            if (_phoneLine.RegState == RegState.RegistrationSucceeded && _call == null)
            {
                _call = _softPhone.CreateCallObject(_phoneLine, dialNumber);
                _call.CallStateChanged += Call_StateChanged;
                _call.Start();
            }
        }

        /// <summary>
        /// Hangup the call.
        /// </summary>
        public virtual void HangUp()
        {
            if (_call != null)
            {
                _call.HangUp();
                _call = null;
            }
        }

        /// <summary>
        /// Accept the incoming call request.
        /// </summary>
        public virtual void Accept()
        {
            if (_call != null)
                _call.Answer();
        }

        /// <summary>
        /// Reject the incoming call request.
        /// </summary>
        public void Reject()
        {
            if (_call != null)
            {
                _call.Reject();
                _call = null;
            }
        }

        public void StreamPublished(IMediaStream mediaStream)
        {
            _mediaGatewayAudioSender = new MediaGatewayAudioSender(mediaStream);
            _mediaConnector.Connect(_mediaGatewayAudioSender, _phoneCallAudioSender);
        }

        public void StreamClosed()
        {
            if (_mediaGatewayAudioSender != null)
            {
                _mediaConnector.Disconnect(_mediaGatewayAudioSender, _phoneCallAudioSender);
                _mediaGatewayAudioSender.Dispose();
                _mediaGatewayAudioSender = null;
            }
        }

        #region VoIP SDK events

        /// <summary>
        /// Indicates phoneline registration changes.
        /// </summary>
        void PhoneLine_RegistrationStateChanged(object sender, RegistrationStateChangedArgs e)
        {
            if (e.State == RegState.RegistrationSucceeded)
            {
                OnSetReadyStatus(true, Account.DisplayName);
                return;
            }

            if (e.State == RegState.Error)
            {
                OnSetReadyStatus(false, e.ReasonPhrase);
                return;
            }

            if (e.State == RegState.NotRegistered)
                OnSetReadyStatus(false, e.State.ToString());
        }

        /// <summary>
        /// Indicates received incoming call.
        /// </summary>
        void SoftPhone_IncomingCall(object sender, VoIPEventArgs<IPhoneCall> e)
        {
            if (e.Item.PhoneLine == _phoneLine)
            {
                _call = e.Item;
                _call.CallStateChanged += Call_StateChanged;

                OnCallRequest(_call.DialInfo.CallerID);
            }
        }

        /// <summary>
        /// Indicates call state changes.
        /// </summary>
        protected virtual void Call_StateChanged(object sender, CallStateChangedArgs e)
        {
            if (e.State == CallState.Answered)
            {
                _phoneCallAudioSender.AttachToCall(_call);
                _phoneCallAudioReceiver.AttachToCall(_call);

                _publishStreamName = Guid.NewGuid().ToString();
                _playStreamName = Guid.NewGuid().ToString();

                var mediaStream = _streamService.CreateStream(_playStreamName);
                _mediaGatewayAudioReceiver = new MediaGatewayAudioReceiver(mediaStream);

                _mediaConnector.Connect(_phoneCallAudioReceiver, _mediaGatewayAudioReceiver);

                OnInCall(_publishStreamName, _playStreamName);
            }

            if (e.State.IsCallEnded())
            {
                _phoneCallAudioSender.Detach();
                _phoneCallAudioReceiver.Detach();

                _streamService.RemoveStream(_mediaGatewayAudioReceiver.MediaStream);
                _mediaConnector.Disconnect(_phoneCallAudioReceiver, _mediaGatewayAudioReceiver);
                _mediaGatewayAudioReceiver.Dispose();

                _mediaGatewayAudioReceiver = null;
                _call = null;
            }

            OnCallStateChanged(e.State);
        }

        #endregion

        #region Client-side method invoke

        /// <summary>
        /// Notifies the client that call state has changed.
        /// </summary>
        void OnCallStateChanged(CallState callState)
        {
            Console.WriteLine("OnCallStateChanged");
            _client.InvokeMethod("OnCallStateChanged", callState);
        }

        /// <summary>
        /// Invokes client's callback method. Notifies the client that it is ready to start or receive a call.
        /// </summary>
        public void OnSetReadyStatus(bool isReady, string name)
        {
            try
            {
                _client.InvokeMethod("OnSetReadyStatus", isReady, name);
            }
            catch (Exception)
            { }
        }

        /// <summary>
        /// Invokes client's callback method. Notifies the client that it received a new call from another extension.
        /// </summary>
        /// <param name="requestOwner">The call request sender's clientId.</param>
        public void OnCallRequest(string requestOwner)
        {
            Console.WriteLine("Call request received from {0} to {1}", requestOwner, "Name");

            _client.InvokeMethod("OnCallRequest", requestOwner);
        }

        /// <summary>
        /// Invoke client's callback method. Sends the publish and receive stream ID.
        /// </summary>
        public void OnInCall(string senderPublishID, string receiverPlayId)
        {
            Console.WriteLine("Sends 'start publishing' sign to the clients.");
            _client.InvokeMethod("OnInCall", senderPublishID, receiverPlayId);

        }

        #endregion
    }
}
